/***************************************************************
/
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

// This file defines all of the functions necessary to dynamically determine an
// optimal number of parallel streams for a globus-url-copy transfer.

#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <stdint.h>
#include <inttypes.h>
#include <unistd.h>

#include "dap_guc_optimization.h"
#include "sqlite3.h"
#define WANT_CLASSAD_NAMESPACE
#include "classad/classad_distribution.h"

#define round(x) (int)(x+0.5)
#define DEFAULT_STREAMS 1
#define MAX_STREAMS 40
#define BUFFER_SIZE 1024
#define HISTORY_DB "opt_history.db"

// Global handle to history database
static sqlite3 *history_db = NULL;

// HELPER FUNCTIONS
// ----------------
// Some small, useful functions which make doing mundane things easy (and fun!)

// Add a sample to a sample collection. Return 0 if the sample has more
// samples than MAX_SAMPLES allows. Otherwise return the number of samples.
int add_sample(uint8_t streams, double throughput, thrp_samples *samples) {
	if (samples->count >= MAX_SAMPLES) return 0;

	//printf("Got sample: (%d, %f)\n", streams, throughput);

	thrp_sample_res *tsr = &samples->res[samples->count];
	tsr->sc = streams;
	tsr->thrp = throughput;
	samples->count++;
	return samples->count;
}

// Convert a string like 100M or 5G into a uintmax_t.
// Supports suffixes: K, M, G, T
uintmax_t filesize_in_bytes(char *fs) {
	char *p;
	uintmax_t tmp;

	tmp = strtoumax(fs, &p, 10);

	switch (*p) {
	  case 't':
	  case 'T':
		tmp *= 1024;
	  case 'g':
	  case 'G':
		tmp *= 1024;
	  case 'm':
	  case 'M':
		tmp *= 1024;
	  case 'k':
	  case 'K':
		tmp *= 1024;
	  default:
		return tmp;
	}
}

// Calculate the difference of two terms n2^2/thn2^2 n1^2/thn1^2 given
// two throughput sample results.
inline double cal_dif(thrp_sample_res tsr1, thrp_sample_res tsr2) {
	int n1 = tsr1.sc, n2 = tsr2.sc;  // Stream counts
	double thn1 = tsr1.thrp, thn2 = tsr2.thrp;  // Throughputs
	return (n1*n1/thn1/thn1 - n2*n2/thn2/thn2) / (n1-n2);
}

// Calculate the parameter a given results array and sample numbers i, j, k.
inline double cal_a(thrp_sample_res *tsr, int i, int j, int k) {
	return (cal_dif(tsr[k], tsr[i]) - cal_dif(tsr[j], tsr[i])) /
	       (tsr[k].sc - tsr[j].sc);
}

// Calculate the parameter b given results array and sample numbers i, j, k.
inline double cal_b(thrp_sample_res *tsr, int i, int j, int k) {
	return cal_dif(tsr[j], tsr[i]) -
	       (tsr[i].sc + tsr[j].sc) * cal_a(tsr, i, j, k);
}

// Calculate the parameter c given results array and sample numbers i, j, k.
inline double cal_c(thrp_sample_res *tsr, int i, int j, int k) {
	uint8_t ni = tsr[i].sc;
	double thni = tsr[i].thrp;
	return ni*ni/thni/thni -
	       ni*ni * cal_a(tsr, i, j, k) -
		   ni * cal_b(tsr, i, j, k);
}

// Calculate the throughput of n streams as predicted by our model.
inline double cal_thr(double a, double b, double c, int n) {
	if (a*n*n + b*n + c <= 0) return 0;
	return n / sqrt(a*n*n + b*n + c);
}

// Calculate the peak (i.e. optimal) stream count based on the prediction model.
inline int cal_full_peak(double a, double b, double c) {
    int n;
    double thr1, thr2;

    // Find the point where the ratio of increase is less than 10%
    for (n = 1; n <= MAX_STREAMS; n++) {
        thr1 = cal_thr(a, b, c, n);
        thr2 = cal_thr(a, b, c, n+1);
        if ((thr2 - thr1) < (thr1 * 0.1)) break;
    }

    // Now just do some bounds checking
    if (n > MAX_STREAMS) n = MAX_STREAMS;
    return n;
}

// Calculate the error when given a, b, c. df is a pointer to the "discard
// flag" which indicates if a value has been discarded from the calculation of
// the error.
inline double cal_err(thrp_sample_res *tsr, int tsr_c,
               double a, double b, double c, int *df) {
	int i, num = tsr_c;
	double thr, sqr_sum = 0.0;

	for (i = 0; i < tsr_c; i++) {
		thr = cal_thr(a, b, c, tsr[i].sc);
		if (thr <= 0)
			num--;
		else
			sqr_sum += thr*thr;
	}

	// Set df if num was decremented by the loop at all.
	*df = num != tsr_c;

	return sqrt(sqr_sum/num);
}

// OUTPUT FUNCTIONS
// ----------------
// Functions used for printing debugging output. Remove these later maybe?
// Return a statically allocated string representing the time required to transfer
// filesize bytes given throughout.
char *est_tran_time(uintmax_t filesize, double throughput) {
	static char buf[BUFFER_SIZE];
	uintmax_t sec = filesize / throughput;
	
	if (throughput > 0.0)
		snprintf(buf, BUFFER_SIZE, "%ju hours, %ju mins, %ju secs",
		         sec/3600, (sec/60)%60, sec%60);
	else
		snprintf(buf, BUFFER_SIZE, "???");
	
	return buf;
}

// Print an estimation about the transfer time given file size
// and peak throughput.
inline void print_est_tran_time(uintmax_t filesize, double throughput) {
	uintmax_t sec;

	printf("Bytes remaining to be transferred:\t [ %ju ]\n"
	       "The estimated maximum throughput:\t [ %f MB/sec ]\n"
	       "The estimated time:\t\t\t ",
	       filesize, throughput/1024/1024);

	if (throughput >= 0.0) {
		sec = filesize/throughput;
		printf("[ %ju hours, %ju mins, %ju secs ]\n",
		       sec/3600, (sec/60)%60, sec%60);
	} else {
		printf("[ ??? ]\n");
	}
}

// OPTIMIZATION FUNCTIONS
// ----------------------
// The big functions which do all the work involved in taking samples and
// coming up with the optimal number of parallel streams.

// Take a sample throughput given the information provided by ts. Return -1 if
// there is a problem (e.g. globus-url-copy fails).
// This function may (read: will likely) alter the contents of ts.
double get_throughput(transfer_status *ts) {
	int status;
	uintmax_t sample, rem;
	char command[BUFFER_SIZE];
	struct timeval tran_start;
	struct timeval tran_finish;

	// Don't bother continuing if there's nothing left to transfer.
	if (ts->off >= ts->size) return 0;
	rem = ts->size - ts->off;  // Remaning file length

	
// Don't pick a sample size bigger than the rest of the file.
	sample = (rem < ts->sample) ? rem : ts->sample;

	// Build GUC command, check that command was generated correctly.
	status = snprintf(command, BUFFER_SIZE,
	                  "globus-url-copy -p %hhu -off %ju -len %ju '%s' '%s'",
	                  ts->streams, ts->off, sample, ts->src_url, ts->dest_url);
	if (status < 0 || status >= BUFFER_SIZE) return -1;

	// Start executing GUC...
	gettimeofday(&tran_start, NULL);

	status = system(command);
	if (status != EXIT_SUCCESS) return -1;

	gettimeofday(&tran_finish, NULL);
	// End executing GUC...

	// Get difference between tran_start and tran_finish.
	tran_finish.tv_sec -= tran_start.tv_sec;
	tran_finish.tv_usec -= tran_start.tv_usec;
	if (tran_finish.tv_usec < 1000000) {
		tran_finish.tv_sec--;
		tran_finish.tv_usec += 1000000;
	}

	// Update ts
	ts->off += sample;  // Move the offset up to end of sample

	// Now if we are doing dynamic optimization, we need to check len...
	// If the sample crossed the chunk boundary, increase len to compensate
	// until it's big enough to transfer the remaining portion of some chunk.
	if (ts->dynamic) {
		rem -= sample;  // Now rem is amount remaining after sampling.

		while (ts->len < sample) {
			// Don't add more than is remaining in the file!
			ts->len += (rem < ts->chunk) ? rem : ts->chunk;
			rem -= ts->chunk;
		} ts->len -= sample;  // Now shave off whatever we used to sample
	}  // A lot of work for such an unlikely situation, huh? Gotta be thorough!

	return sample / (tran_finish.tv_sec + tran_finish.tv_usec/1000000.0);
}

// Take a number of samples at different stream nums and record throughputs
// into the thrp_samples container pointed to by tps.
void do_sampling(transfer_status *ts, thrp_samples *tps) {
	uint8_t str1, str2;  // Number of streams
	double thr0, thr1, thr2;
	double alpha = 0.001;

	// Test with default number of streams.
	ts->streams = str1 = DEFAULT_STREAMS;
	thr0 = thr1 = get_throughput(ts);
	if (thr1 <= 0) return;
	if (!add_sample(str1, thr1, tps)) return;

	// Then with two.
	ts->streams = str2 = 2;
	thr2 = get_throughput(ts);
	if (thr2 <= 0) return;
	if (!add_sample(str2, thr2, tps)) return;



////////////////

/*while ((thr2-thr1)>(thr1*0.1))
{
printf("inside while loop thr1 and thr2: %3.2f, %3.2f",thr1, thr2);
str1=str2;
thr1=thr2;
ts->streams = str2*=2;
thr2 = get_throughput(ts); 
            if (thr2 <= 0) return;
          if (!add_sample(str2, thr2, tps)) return;

}
	// Loop until we reach the optimal number of streams.
*/
	while ((thr2-thr1) / ((str2-str1)*thr0) > alpha) {
		str1 = str2;
		thr1 = thr2;
		ts->streams = str2 *= 2;
		thr2 = get_throughput(ts);
		if (thr2 <= 0) return;
		if (!add_sample(str2, thr2, tps)) return;
	}
}

// Do an analysis on the data contained in tps to determine the optimal number
// of streams, and return the result. Copy the peak throughput into *peak_thrp.
int do_analysis(thrp_samples *tps, double *peak_thrp, double *single_thrp) {
	thrp_sample_res *tsr = tps->res;  // Array of throughput sample results
	int tsr_c = tps->count;  // Throughput sample result count (# of samples)
	int ip = 0, jp = 0, kp = 0;  // Optimal i, j, and k.
	double a, b, c, err = INFINITY;  // Optimal a, b, c, and err.
	int i, j, k;  // i, j, and k used by the loop.
	int df;  // Discard flag set by cal_err indicating a datum was ignored.
	int peak_sc;
	// We can only do analysis if there are at least three samples
	if (tsr_c < 3) {
		printf("Not enough data points collected to do analysis\n");
		peak_sc = DEFAULT_STREAMS;
		// Use the first throughput if there is one, otherwise 0.
		*peak_thrp = (tsr_c > 0) ? tsr[0].thrp : 0;
		goto exit;
	}

	// Iterate through the samples, find the "best" three.
	for (i = 0;   i < tsr_c-2; i++)
	for (j = i+1; j < tsr_c-1; j++)
	for (k = j+1; k < tsr_c;   k++) {
		double a_, b_, c_, err_;
		a_ = cal_a(tsr, i, j, k);
		b_ = cal_b(tsr, i, j, k);
		c_ = cal_c(tsr, i, j, k);
		err_ = cal_err(tsr, tsr_c, a_, b_, c_, &df);
		
		// If new err is better, replace old calculations.
		if (err_ < err && !df) {
			err = err_;
			a = a_, b = b_, c = c_;
			ip = i, jp = j, kp = k;
		}
	}

	*single_thrp = cal_thr(a, b, c, 1);



	printf("Chosen (stream, throughput) pairs are:\n"
	       "(%d, %3.10f), (%d, %3.10f), (%d, %3.10f)\n",
	       tsr[ip].sc, tsr[ip].thrp,
	       tsr[jp].sc, tsr[jp].thrp,
	       tsr[kp].sc, tsr[kp].thrp);
	printf("The value of a, b, c, and err:\n"
	       "a = %g, b = %g, c = %g, err = %g)\n",
	       a, b, c, err);

	// Calculate peak stream count and throughput.
	peak_sc = cal_full_peak(a, b, c);
	*peak_thrp = cal_thr(a, b, c, peak_sc);

	exit:
	return peak_sc;
}

// HISTORY DATABASE FUNCTIONS
// --------------------------
// Functions used to create and add/update entires in the optimization
// history database.

// Open the database for reading and writing, store handle to history_db if
// successful.
int open_history_db() {
	sqlite3 *db;

	if (sqlite3_open(HISTORY_DB, &db)) {
		fprintf(stderr, "Cannot open history database: %s\n", sqlite3_errmsg(db));
		sqlite3_close(db);
		return 0;
	}

	// Close the old connection if we've already opened one
	if (history_db) sqlite3_close(db);

	history_db = db;
	return 1;
}

// What do you think it does?
void close_history_db() {
	if (history_db) sqlite3_close(history_db);
	history_db = NULL;
}

// Read history from the database, store results to ts and optional arguments.
int get_history(transfer_status *ts, int *o_sc, double *o_thrp) {
	char *src, *dest;
	int ret;  // Return values
	char sql[BUFFER_SIZE];  // Holds the SQL statement
	static sqlite3_stmt *p_stmt = NULL;  // Prepared statement handle

	// Get src and dest from ts
	src = ts->src_url;
	dest = ts->dest_url;

	// Prepare SQL statement if it's not already been.
	if (!p_stmt) {
		snprintf(sql, BUFFER_SIZE,
			"SELECT sc, thrp FROM opt_info"
			"	WHERE src = :src AND dest = :dest"
			"	LIMIT 1");
		ret = sqlite3_prepare_v2(history_db, sql, BUFFER_SIZE, &p_stmt, NULL);
		if (ret != SQLITE_OK) return 0;
	} else {
		sqlite3_reset(p_stmt);
	}

	// Bind parameters
	sqlite3_bind_text(p_stmt, 1, src, strlen(src), NULL);
	sqlite3_bind_text(p_stmt, 2, dest, strlen(dest), NULL);

	// Execute the statement and parse results
	if (sqlite3_step(p_stmt) == SQLITE_ROW) {
		if (o_sc) *o_sc = sqlite3_column_int(p_stmt, 0);
		if (o_thrp) *o_thrp = sqlite3_column_double(p_stmt, 1);
		printf("History for %s and %s:\n\tsc = %d, thrp = %f\n", src, dest,
		       sqlite3_column_int(p_stmt, 0), sqlite3_column_double(p_stmt, 1));
		return 1;
	} else {  // Error or nothing found
		printf("No history for %s and %s.\n", src, dest);
		return 0;
	}
}

// Add history information to the database specified by history_db global.
int add_history(transfer_status *ts, int sc, double thrp) {
	char *src, *dest;
	int ret;  // Return values
	int retry = 10;  // Number of times to retry if there is a lock
	static bool table_created = false;
	char sql[BUFFER_SIZE];  // Holds the SQL statement
	char *errmsg = NULL;  // Free by calling sqlite3_free
	static sqlite3_stmt *p_stmt = NULL;  // Prepared statement handle

	// Get src and dest from ts
	src = ts->src_url;
	dest = ts->dest_url;

	// Try to create table if we don't know it exists.
	if (!table_created) {
		snprintf(sql, BUFFER_SIZE,
			"CREATE TABLE IF NOT EXISTS opt_info(src, dest, sc, thrp, date)");
		// Execute statement
		ret = sqlite3_exec(history_db, sql, NULL, NULL, &errmsg);
		if (ret != SQLITE_OK) {  // Error has occurred.
			fprintf(stderr, "Error creating table: %d: %s\n", ret, errmsg);
			sqlite3_free(errmsg);
			return 0;
		} table_created = true;
	}

	printf("Adding history for %s and %s.\n", src, dest);

	// Generate statement to update/insert information and bind values.
	// TODO(bross) Fix this undesireable solution to the problem of generating
	// a working UPSERT statement.
	if (get_history(ts, NULL, NULL)) {
		printf("UPDATING\n");
		snprintf(sql, BUFFER_SIZE,
			"UPDATE opt_info"
			"  SET src = :src, dest = :dest,"
			"      sc = :sc, thrp = :thrp, date = datetime('now','localtime')"
			"  WHERE src = :src AND dest = :dest;");
	} else {
		printf("INSERTING\n");
		snprintf(sql, BUFFER_SIZE,
			"INSERT INTO opt_info VALUES"
			"  (:src, :dest, :sc, :thrp, datetime('now','localtime'));");
	}

	// Prepare it
	ret = sqlite3_prepare_v2(history_db, sql, BUFFER_SIZE, &p_stmt, NULL);
	if (ret != SQLITE_OK) {
		printf("Could not prepare. Got error: %d\n", ret);
		return 0;
	}

	// Bind parameters
	sqlite3_bind_text(p_stmt,   1, src, strlen(src), NULL);
	sqlite3_bind_text(p_stmt,   2, dest, strlen(dest), NULL);
	sqlite3_bind_int(p_stmt,    3, sc);
	sqlite3_bind_double(p_stmt, 4, thrp);

	// Execute the statement
	printf("About to be in the switch.\n");
	do switch (sqlite3_step(p_stmt)) {
	  // All done
	  case SQLITE_DONE:
		printf("SQLITE_DONE\n");
		sqlite3_finalize(p_stmt);
		return 1;
	  // Locks (limited retries)
	  case SQLITE_BUSY:
	  case SQLITE_LOCKED: if (retry-- > 0) continue;
	  // Any other errors
	  default:
		printf("SQLITE_ERROR\n");
		sqlite3_finalize(p_stmt);
		return 0;
	} while (1);
}

// THE "MAIN" FUNCTION
// -------------------
// Take samples of a globus-url-copy transfer with different numbers of
// parallel streams and calculate the optimal number of parallel streams.
// Updates ts to reflect the optimal number of streams.
//
int do_optimization_service(transfer_status *ts) {
       int opt;  // Optimal stream count
       char *est_time, *dap_type, *est_time1;
       clock_t start, end;
       thrp_samples thrs;  // Throughput samples
       thrs.count = 0;
       double cost, peak_thrp, single_thrp;  // Peak throughput set by full_analysis
	time_t t1,t2;

	if (ts->history) {
		open_history_db();
		if (get_history(ts, &opt, &peak_thrp)) {
			printf("Using historical optimum...\n");
			ts->streams = opt;
			return opt;
		} else {
			printf("Histortical optimum not found.\n");
		}
	} else {
		printf("Not checking for histortical maximum.\n");
	}
	
       dap_type = getenv("DAP_TYPE");
       (void) time(&t1);
     //  printf("Starting sampling...\n");
       do_sampling(ts, &thrs);
     //  printf("Starting analysis...\n");
       opt = do_analysis(&thrs, &peak_thrp, &single_thrp);
       ts->streams = opt;
       (void) time (&t2);
      cost = difftime(t2,t1);
		uintmax_t sizes;
       // Debugging output
       printf("The cost of optimization:\t\t [ %.3f secs ]\n", cost);
       printf("Suggested parallel stream number:\t [ %d ]\n", opt);
       printf("ts.size = %ju\n", ts->size);
       est_time =strdup( est_tran_time(ts->size, peak_thrp));
       print_est_tran_time(ts->size - ts->off, peak_thrp);
       est_time1 = est_tran_time(ts->size, single_thrp);

	// Add our findings to the database
	if (strcmp(dap_type, "estimation")) {	
		if (!history_db) open_history_db();
		printf("Adding results to history db.\n");
		add_history(ts, opt, peak_thrp);
	}

	// If we're doing estimation, write output file
       if (dap_type && !strcmp(dap_type, "estimation")) {
	
		write_classad_optimization(opt, peak_thrp, cost, est_time,ts->size,est_time1);
	}
	
	return opt;
}
// This should only be called outside of this file.
// ClassAd stuff
void write_classad_optimization(int opt1, double thrp, double difftime,
                                char *est_time, uintmax_t filesize, char*est_time1) {
	FILE *fp;
	char *path, *str_dap_id;
	classad::ClassAd *classad = new classad::ClassAd;
	classad::ClassAdParser parser;
	
	if (!(str_dap_id = getenv("STORK_JOBID"))) return; 



	// Insert attributes into ClassAd
        classad->InsertAttr("dap_id", atoi(str_dap_id));
        classad->InsertAttr("optimal_streams", opt1);
        classad->InsertAttr("optimization_cost_in_seconds", int ( difftime));
        classad->InsertAttr("est_time_using_optimal_streams", est_time);
        classad->InsertAttr("est_time_using_single_stream", est_time1);

        classad->InsertAttr("max_throughput",thrp*8/1024/1024);
        classad->InsertAttr("file_size_in_megabytes",  int (filesize/1024/1024));

	// Unparse the ClassAd
	classad::PrettyPrint unparser;
	std::string adbuffer = "";
	unparser.Unparse(adbuffer, classad);
	
	delete classad;
	asprintf(&path, "/tmp/%d.log", atoi(str_dap_id));					
	
	if (!(fp = fopen(path, "w"))) {
		printf("cannot open logfile: %s\n", path);
		exit(1);
	}

	fprintf (fp, "%s", adbuffer.c_str());
free(path);
	fclose(fp);
}
