/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

// This file contains all of the code relevant to the anonymous stat
// reporting feature. By default, this feature is enabled. The report
// is sent via an HTTP GET request to the server specified by
// STAT_REPORT_SERVER in the config file (or the hardcoded default
// server; see below), and contains the following information:
//
//   1) the stork_server version string
//   2) the platform string

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "condor_common.h"
#include "condor_config.h"
#include "dap_utility.h"
#include "dap_server.h"

// Stat reports are sent by default to this server
#define DEFAULT_STAT_SERVER "stork.cct.lsu.edu"

// Check if stat reporting is enabled. Defaults to false.
bool stat_report_enabled() {
	return param_boolean("ENABLE_STAT_REPORT", false);
}

// Return the (statically allocated) hostname of the stat server according to
// the config file
char *stat_server_host() {
	static char *host = NULL;
	if (!host) host = param("STAT_REPORT_SERVER");
	if (!host) host = DEFAULT_STAT_SERVER;
	return host;
}

// Open an HTTP connection to the stat server, given the address. Return
// a socket descriptor or -1 if an error occurred. This socket needs to be
// closed when we're done with it.
int connect_to_stat_server(char *stat_server) {
	int s;
	struct addrinfo hints, *res = NULL;

	// Initialize "hints" struct
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	// Create addrinfo struct for the stat server
	if (getaddrinfo(stat_server, "80", &hints, &res)) return -1;

	// Actually open the socket. If we can't, goto fail to clean up.
	s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (s == -1 || connect(s, res->ai_addr, res->ai_addrlen) == -1)
		goto fail;

	freeaddrinfo(res);
	return s;

	// Error handling
	fail:
	if (res) freeaddrinfo(res);
	return -1;
}

// Construct a message to send to the stat report server, given the version
// and platform strings. This should sanitize inputs for HTTP to a
// reasonable degree. The resulting string needs to be freed later.
// Return NULL if there is an error.
char *construct_stat_report(const char* version, const char* platform) {
	char *msg = NULL, *fmtstr, *stat_server, *verstr = NULL, *platstr = NULL;

	fmtstr = "GET http://%s/report.php?version=%s&platform=%s HTTP/1.0\n\n";
	stat_server = stat_server_host();

	// strdup() version and platform so we can modify them
	verstr = strdup(version);
	platstr = strdup(platform);
	if (!verstr || !platstr) goto fail;

	// Replace spaces with + characters. This isn't confusing, is it?
	for (char *p = verstr;  *p != '\0'; p++) if (*p == ' ') *p = '+';
	for (char *p = platstr; *p != '\0'; p++) if (*p == ' ') *p = '+';

	asprintf(&msg, fmtstr, stat_server, verstr, platstr);

	// Now free some stuff and return our report
	free(verstr);
	free(platstr);
	return msg;

	// Error handling
	fail:
	free(verstr);
	free(platstr);
	free(msg);
	return NULL;
}

// Connect to the stat server and issue GET request containing report.
// Returns false if there is a problem, true otherwise.
bool send_stat_report() {
	if (!stat_report_enabled()) return true;

	int sockfd; // Socket descriptor
	char *msg; // This will be the HTTP message we send to the server
	
	// Open an HTTP connection to the server. Close this later.
	sockfd = connect_to_stat_server(stat_server_host());
	if (sockfd == -1) goto fail;

	// Generate the message to give the server. Free this later.
	msg = construct_stat_report(StorkVersion(), StorkPlatform());
	if (!msg) goto fail;

	// Send the message
	sendall(sockfd, msg, strlen(msg), 0);

	free(msg);
	close(sockfd);
	return true;

	// Error handling
	fail:
	free(msg);
	if (sockfd != -1) close(sockfd);
	perror("send_stat_report");
	return false;
}
