/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

#include "dap_stream_list.h"

// XXX: This may become deprecated at some point!
//
// This is a *dumb workaround* until we decide on either scrapping the Condor
// stream library and writing our own, or getting it to work how we'd like it
// to. A stream_list represents a mixed list of C library file handles and
// Condor ReliSocks. The functions in this file should pretty much only be used
// by userlog writing functions in other files, and this file will probably be
// gone in a future release. Therefore, using it elsewhere is not suggested.

// A stream_list allows: appending streams, removing streams, and "broadcasting"
// a string or formatted string to all streams in the list. A stream is either a
// file or a Condor ReliSock. Of course, if we were using our own stream library
// there would be no need to differentiate between the two and we probably
// wouldn't even really need this file. But oh well.
//
// A stream_list also contains a temporary "backing" file which contains all
// data written to the stream list thus far. When a new stream is registered, it
// will automatically write the contents of the backing file to the stream, thus
// bringing it up-to-date.

// Dynamically allocate a new stream_node
stream_node *new_stream_node(void *obj, int type) {
	if (!obj) return NULL;

	stream_node *sn = (stream_node*) malloc(sizeof(stream_node));
	if (!sn) return NULL;

	sn->type = type;
	sn->obj = obj;
	sn->prev = sn->next = NULL;
	return sn;
}

// Dynamically allocate and initialize a new stream_list. Return NULL on error.
stream_list *new_stream_list() {
	stream_list *sl = (stream_list*) calloc(1, sizeof(stream_list));

	if (!sl) return NULL;

	// Create a temporary backing file for the stream. Should we fail on error?
	sl->tmplog = tempnam(NULL, NULL);
	if (!sl->tmplog)
		dprintf(D_ALWAYS, "WARNING: Couldn't create backing log file!\n");

	return sl;
}

// Iterate through a stream_list, close streams, and free all the nodes
void sl_close(stream_list *sl) {
	if (!sl) return;

	// Close and delete backing file
	if (sl->tmplog) {
		remove(sl->tmplog);
		free(sl->tmplog);
	}

	for (stream_node *next_sn, *sn = sl->head; sn; sn = next_sn) {
		// Close associated stream
		switch (sn->type) {
			case SL_FILE_TYPE: {
				FILE *fp = (FILE*) sn->obj;
				fclose(fp);
				break;
			} case SL_SOCK_TYPE: {
				ReliSock *sock = (ReliSock*) sn->obj;
				sock->encode();
				sock->put(DAP_COMPLETE);
				sock->eom();
				sock->close();
				break;
			}
		}

		next_sn = sn->next;  // Best get this before freeing
		free(sn);
	}

	sl->head = sl->tail = NULL;
	free(sl);
}

// List manipulation functions
// ---------------------------
// Add a new stream_node to the list
bool sl_add(stream_list *sl, stream_node *sn) {
	if (!sl || !sn) return false;  // No NULLs allowed!

	if (sl->tail) {  // There's something in the list
		sl->tail->next = sn;  // Append to old tail
		sl->tail = sn;        // Set new tail
	} else sl->head = sl->tail = sn;

	// FIXME: For some reason this is needed and no one knows why. :(
	sl_node_write(sn, (void*)"BLAH!", 5);

	// Copy the contents of the backing file to the new stream
	if (sl->tmplog) {
		char buf[256];
		int len;
		FILE *fp = fopen(sl->tmplog, "rb");
		while (len = fread(buf, 1, sizeof(buf), fp))
			sl_node_write(sn, buf, len);
		fclose(fp);
		dprintf(D_ALWAYS, "Closing tmplog\n");
	}

	return true;
}

// Create a new stream_node with the specified type and add it
bool sl_add(stream_list *sl, void *obj, int type) {
	return sl_add(sl, new_stream_node(obj, type));
}

// Add a file
bool sl_add(stream_list *sl, FILE *fp) {
	return sl_add(sl, fp, SL_FILE_TYPE);
}

// Add a socket
bool sl_add(stream_list *sl, ReliSock *sock) {
	sock->encode();
	sock->put(DAP_SUCCESS);
	sock->eom();
	return sl_add(sl, sock, SL_SOCK_TYPE);
}

// Remove all references to something from the list. This is kinda inefficient,
// but oh well.
void sl_del(stream_list *sl, void *obj) {
	if (!sl || !obj) return;

	for (stream_node *sn = sl->head; sn; sn = sn->next) if (sn->obj == obj) {
		if (sn == sl->head) sl->head = sn->next;
		if (sn == sl->tail) sl->tail = sn->prev;
		if (sn->prev) sn->prev->next = sn->next;
		if (sn->next) sn->next->prev = sn->prev;
		free(sn);
	}
}

// List broadcasting functions
// ---------------------------
// Write a formatted string to every stream in the list
void sl_printf(stream_list *sl, char *fmt, ...) {
	va_list ap;
	char *str = NULL;
	int len;

	va_start(ap, fmt);

	// Allocate a new string holding the expanded format string
	len = vasprintf(&str, fmt, ap);
	if (!str) return;

	sl_write(sl, str, len);

	free(str);
	va_end(ap);
}

// Write a null terminated string to every stream in the list
void sl_put(stream_list *sl, char *str) {
	sl_write(sl, str, strlen(str));
}

// Write a block of bytes to every stream in the list
void sl_write(stream_list *sl, void *bytes, int len) {
	if (!sl || !bytes || len <= 0) return;

	// First, write to backing file and close it
	if (sl->tmplog) {
		FILE *fp = fopen(sl->tmplog, "ab");
		fwrite(bytes, 1, len, fp);
		fclose(fp);
	}

	// Iterate and write to nodes on list
	for (stream_node *sn = sl->head; sn; sn = sn->next)
		sl_node_write(sn, bytes, len);
}

// Write a block of bytes to a stream_node
void sl_node_write(stream_node *sn, void *bytes, int len) {
	if (!sn || !bytes || len <= 0) return;

	switch (sn->type) {
		case SL_FILE_TYPE: {  // Handle writing to files
			FILE *fp = (FILE*) sn->obj;
			fwrite(bytes, 1, len, fp);
			fflush(fp);
			break;
		} case SL_SOCK_TYPE: {  // Handle writing to sockets
			ReliSock *sock = (ReliSock*) sn->obj;
			sock->encode();
			sock->put(DAP_SUCCESS);
			sock->put_bytes(bytes, len);
			sock->put('\0');
			sock->eom();
			break;
		} default:  // Uhh, how did this happen
			dprintf(D_ALWAYS, "Somewhere stream_list is being abused.\n");
	}
}
