/***************************************************************
*
* Copyright (C) 2008-2010, Distributed Systems Laboratory,
* Center for Computation & Technology, and Department of Computer Science
* Louisiana State University, Baton Rouge, LA.
*
* Licensed under the Apache License, Version 2.0 (the "License"); you
* may not use this file except in compliance with the License.  You may
* obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*
***************************************************************/

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include <time.h>
#include <pthread.h>

#include "condor_common.h"
#include "dap_utility.h"
#include "dap_error.h"
#include "dap_constants.h"
#include "dap_server_interface.h"
#include "dap_client_interface.h"
#include "condor_config.h"

// Debugging output
#define DPRINT(...) fprintf(stderr, __VA_ARGS__)

// Names for certain files
#define PID_FILE_NAME "stork_log.pid"
#define SOCK_PREFIX "stork_log.sock."

// Minimum arguments for commands
#define MIN_GET_ARGS  3
#define MIN_LIST_ARGS 0

// Maximum message length for client -> server messages
#define MAX_MSG_LEN 1024
#define LOG_BUF_LEN 1024

// Seconds a daemon will stay alive by default
#define DEFAULT_DAEMON_LIFE 30

// Usage information for this program
#define USAGE \
"Usage: stork_log [-h|--help] [-p pid] <command> [args] ...\n" \
"\n" \
"A tool to transfer log files for Stork jobs from a Stork server to the\n" \
"local machine.\n" \
"\n" \
"Options:\n" \
"  -h, --help  Print this usage information and exit\n" \
"  -p <pid>    Specify the pid of the stork_log daemon to issue commands to\n" \
"  -q          Don't print anything to stdout\n" \
"\n" \
"Valid commands and associated parameters are:\n" \
"  get <host[:port]> <dap_id> <file_path>\n" \
"    Retrieve a userlog for job with given dap_id from the Stork server at\n" \
"    host[:port], and place it at file_path.\n" \
"\n" \
"  list [dap_id]\n" \
"    If dap_id is given, show the status of log transfers for that dap_id.\n" \
"    Otherwise, show the status of all ongoing log transfers.\n" \
"    (NOT YET IMPLEMENTED)\n"

// Forward declarations
void usage_and_exit(int, char*);
char *pid_path();
char *sock_path();
char *sock_path(int);
struct dap_log_node;

// stork_log should have two uses:
//   1) Run from the command line, shows status of stork_log transfers.
//   2) Run as a daemon, transfers Stork logs from the stork_server to the
//      local file system, and accepts requests for such transfers.

// Supported commands
// ------------------
// stork_log in daemon mode supports only one command to retrieve a log file
// from a given Stork server.
//
// GET <server_name> <dap_id> <file_path>
//   Starts a log transfer between the local machine and the Stork server.
//
//   server_name: the address(:port) of the Stork server
//   dap_id: the job ID of the job whose log is being requested
//   file_path: path on the local filesystem where the log should be written
//
// LIST <dap_id>
//   List all transfers and their progress for dap_id. If dap_id is zero, show
//   all transfers. (Not yet implemented.)

// Struct to maintain the state of a stork_log transfer
struct dap_log_xfer {
	unsigned int dap_id;
	char *file_path;
	uintmax_t xferred;  // How much has been transferred
	int status;  // Indicates the status of the transfer
	dap_log_xfer *prev, *next;  // Previous and next transfers in the list
};

// State variable for global linked list of transfer structs
unsigned int xfer_count;
dap_log_xfer *head, *tail;

// Operations for implementing a doubly-linked list of dap_log_xfers. These
// operations NEED to be atomic. TODO: Actually implement this.
void append_job(dap_log_xfer *dlx) { }

// SERVER OPERATIONS
// -----------------
// Functions related to operating stork_log as a daemon to transfer logs.

// Start listening on a Unix socket with address specified by path. Return
// socket descriptor or -1 on error.
int listen_on_unix_socket(char *path) {
	int sd;
	socklen_t len;
	sockaddr_un local;

	// Create socket descriptor
	sd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sd == -1) return -1;

	// Prepare local struct and sock path
	local.sun_family = AF_UNIX;
	sprintf(local.sun_path, "%s", path);
	unlink(local.sun_path);

	// Bind the socket to the address and start listening
	len = strlen(local.sun_path) + sizeof(local.sun_family);
	if (bind(sd, (sockaddr*) &local, len) == -1) return -1;
	if (listen(sd, 10) == -1) return -1;

	return sd;
}

// Create pid file so the daemon can be more easily located. If the file
// already exists, read the pid from it, and check if the process is running.
// If so, tell the user and don't create the pid file. If not, just overwrite
// the file. Returns 1 if it created a file, 0 if there was a problem.
int create_pid_file(char *path) {
	FILE *fp;
	
	printf("Trying to create pid file: %s\n", path);

	// Check if file exists and, if so, if daemon is running
	fp = fopen(path, "w");
	if (!fp) {
		printf("A stork_log daemon is already running! PID file not created.\n");
		return 0;
	}

	fprintf(fp, "%d\n", (int) getpid());

	// Clean up
	fclose(fp);
	return 1;
}

// Handle GET requests, given (s)erver, (d)ap id, and file (p)ath.
int handle_get_request(int rem_sd, char *s, int d, char *p) {
	FILE *fp;
	int ret;
	char *err_msg;
	ReliSock *sock = NULL;

	// Check for null/zero/empty string, because we're paranoid
	if (!(d && p && s && *p && *s)) {
		printf("A bad argument was passed to handle_get_request.\n");
		return DAP_ERROR;
	}

	// Open p for appending
	fp = fopen(p, "a");
	if (!fp) {
		printf("%s could not be opened for writing.\n", p);
		return DAP_ERROR;
	}

	// FIXME: Stupid hack. If "default" was given for the server name, pass NULL
	// to init_stork_log so it knows to use the default server.
	if (!strcmp(s, "default")) s = NULL;

	// Add log transfer to global transfer list
	//register_log_transfer();

	// Initialize the transfer
	ret = init_stork_log(s, d, (Sock**) &sock, &err_msg);
	if (ret != DAP_SUCCESS) {
		printf("Log transfer could not be inited: %s\n", err_msg);
		free(err_msg);
		goto fail;
	}
	
	// Inform the client of success
	send(rem_sd, &ret, sizeof(ret), 0);

	// Now transfer the whole rest of the userlog
	ret = read_stork_log(sock, fp, &err_msg);
	if (ret != DAP_SUCCESS) {
		printf("Log transfer failed: %s\n", err_msg);
		free(err_msg);
		goto fail;
	}

	// Remove log transfer from transfer list

	// Clean up
	fclose(fp);
	free(err_msg);
	sock->close();

	return DAP_SUCCESS;

	// Error handling
	fail:
	if (fp) fclose(fp);
	if (sock) sock->close();

	ret = DAP_ERROR;
	send(rem_sd, &ret, sizeof(ret), 0);

	return DAP_ERROR;
}

// Handle LIST requests. Pass dap_id = 0 to list all.
int handle_list_request(unsigned int dap_id) {
	// Not yet implemented, roflcopters.
	return DAP_ERROR;
}

// Interpret and execute a command received from a client. The return value of
// this function is passed back to the client over the socket. Presumably the
// calling function doesn't care if we mess with its buffer.
int parse_command(int rem_sd, char *cmd) {
	int ret = DAP_ERROR;  // Return value
	char *cmd_type, *delim = " \t", *last = NULL;

	// Some basic checks -- no null or empty strings
	if (!cmd || !*cmd) return DAP_ERROR;

	// Parse the cmd_type out of the command string
	cmd_type = strtok_r(cmd, delim, &last);

	// Handle request, or else whine about unsupported command
	if (!strcmp(cmd_type, "GET")) {
		DPRINT("In GET parsing code\n");
		// Parse out GET parameters
		char *server_name, *dap_id, *file_path;
		int dap_id_i;
		server_name = strtok_r(NULL, delim, &last);
		dap_id =      strtok_r(NULL, delim, &last);
		file_path =   strtok_r(NULL, delim, &last);

		// Check for NULLs
		if (!(file_path && dap_id && server_name)) {
			fprintf(stderr, "Not enough args for GET\n");
			return DAP_ERROR;
		}

		// Check that arguments are valid
		if ((dap_id_i = atoi(dap_id)) <= 0) {  // Positive dap_id
			fprintf(stderr, "dap_id must be a positive integer\n");
			return DAP_ERROR;
		}

		// We should also make sure the file path is writable, server is
		// reachable, etc. Later though.

		// Pass the arguments to the GET request handler and return
		return handle_get_request(rem_sd, server_name, dap_id_i, file_path);
	} else if (!strcmp(cmd_type, "LIST")) {
		fprintf(stderr, "Not yet implemented: %s\n", cmd_type);
		return DAP_ERROR;
	} else if (!strcmp(cmd_type, "NOP")) {
		// Don't even do anything
		return DAP_SUCCESS;
	} else {  // We don't support that! :(
		fprintf(stderr, "Unsupported command: %s\n", cmd_type);
		return DAP_ERROR;
	}
}

// Entry point for threads spawned by the daemon to handle parsing of client
// commands.
void *handle_client_conn(void *arg) {
	int ret, len;
	int rem_sd = *((int*) arg);  // Remote socket descriptor
	char *buf;

	xfer_count++;

	// Read command from client
	buf = (char*) malloc(MAX_MSG_LEN+1);
	if (!buf) goto fail;
	switch(len = recv(rem_sd, buf, MAX_MSG_LEN, 0)) {
		case 0: printf("client closed connection\n");
		case -1: goto fail;
		default: buf[len] = '\0';
	}
	
	// Parse and handle command
	parse_command(rem_sd, buf);

	if (xfer_count) xfer_count--;

	free(buf);

	pthread_exit(NULL);
	return NULL;

	// Exit
	fail:
	perror("handle_client_conn");
	free(buf);
	if (xfer_count) xfer_count--;
	pthread_exit(NULL);
	return NULL;
}

// Clean up when the program exits from daemon mode
void daemon_cleanup() {
	// Unlink 
	unlink(sock_path());
	unlink(pid_path());  // Should only be done if we created a pid file
}

// Handle signals. Right now, just exit.
void sig_hndlr(int) { exit(-1); }

// Kill the daemon after it has been alive for t seconds without any jobs
void *daemon_reaper(void *time_p) {
	int death_time;
	int dt;

	if (!time_p || (dt = *((int*)time_p)) <= 0) return NULL;

	reset:
	while (xfer_count) sleep(1);  // Wait until there are no transfers

	// Set the death time
	death_time = dt;

	// Every second, check the time and xfer_count
	while (--death_time > 0) {
		if (xfer_count) goto reset;  // A transfer was started
		sleep(1);
	}

	// When we make it here, it's all over
	exit(EXIT_SUCCESS);
	return NULL;
}

// Begin listening for log transfer requests. Wait for connections on a socket,
// and pass the remote socket descriptor to a new thread to handle servicing
// the client. If more than daemon_life seconds pass without any active
// transfers, the daemon should die. (Or if daemon_life is <= 0, don't.)
int daemon_mode(int daemon_life = DEFAULT_DAEMON_LIFE) {
	int loc_sd = -1, rem_sd = -1, _errno;
	pthread_t t_id, reaper_id;
	char *path;

	// Set up signal handlers
	signal(SIGTERM, sig_hndlr);
	signal(SIGINT,  sig_hndlr);
	signal(SIGQUIT, sig_hndlr);
	signal(SIGHUP,  sig_hndlr);
	atexit(daemon_cleanup);

	// Start listening for clients
	path = sock_path();
	loc_sd = listen_on_unix_socket(path);
	if (loc_sd < 0) goto fail;

	// Create pid file
	path = pid_path();
	if (!path) goto fail;
	create_pid_file(path);

	// Initialize xfer_count semaphore and spawn reaper thread
	if (daemon_life > 0)
		pthread_create(&reaper_id, NULL, daemon_reaper, &daemon_life);

	// Wait for connections and handle them
	for (;;) {
		printf("Waiting for connections...\n");
		rem_sd = accept(loc_sd, NULL, NULL);

		if (rem_sd != -1) {
			pthread_create(&t_id, NULL, handle_client_conn, &rem_sd);
		} else {
			perror("daemon_mode");
		}
	}

	close(loc_sd);
	exit(EXIT_SUCCESS);

	// Error handling
	fail:
	_errno = errno;
	if (loc_sd != -1) close(loc_sd);
	if (rem_sd != -1) close(rem_sd);
	errno = _errno;
	perror("daemon_mode");
	exit(EXIT_FAILURE);
}

// CLIENT OPERATIONS
// -----------------
// Functions related to communicating with a stork_log daemon to retrieve
// information or control the daemon.

// Get the PID of the server from PID file, return 0 if not found
int get_server_pid() {
	FILE *f = NULL;
	int pid = 0;

	// Open the file and read the PID
	f = fopen(pid_path(), "r");
	if (!f) goto fail;
	fscanf(f, "%d", &pid);
	if (pid <= 0) goto fail;

	// Clean up
	fclose(f);
	return pid;

	fail:
	if (f) fclose(f);
	return 0;
}

// Send a command string to the server given sock_fd, return response code or
// -1 if a communication error occured.
int send_command(int sock_fd, char *cmd) {
	int len, resp = DAP_ERROR;

	DPRINT("In send_command\n");

	// Send the command
	len = strlen(cmd);
	if (sendall(sock_fd, cmd, len, 0) != len) return -1;

	DPRINT("Still in send_command\n");

	// Receive and return the response
	switch (recv(sock_fd, &resp, sizeof(resp), 0)) {
		case 0: errno = ECONNRESET;
		case -1:
		perror("send_command");
		DPRINT("There was an error!\n");
		return -1;
	}
	
	DPRINT("Response is: %d\n", resp);

	return resp;
}

// Connect to a daemon with given pid and send it a request, returning socket
// descriptor or exiting on failure.
int connect_to_daemon(int pid) {
	int s = -1, len, _errno;
	sockaddr_un remote;

	if (pid <= 0) {
		errno = EINVAL;
		goto fail;
	}

	// Create socket descriptor
	s = socket(AF_UNIX, SOCK_STREAM, 0);
	if (s == -1) goto fail;

	// Prepare remote struct and connect
	remote.sun_family = AF_UNIX;
	sprintf(remote.sun_path, "%s", sock_path(pid));
	len = strlen(remote.sun_path) + sizeof(remote.sun_family);
	if (connect(s, (sockaddr*) &remote, len) == -1) goto fail;

	// Return socket descriptor
	printf("Connected!\n");
	return s;

	fail:
	_errno = errno;
	if (s != -1) close(s);
	errno = _errno;
	perror("connect_to_daemon");
	exit(EXIT_FAILURE);
}

// CLIENT COMMAND HANDLERS
// -----------------------
// Handlers for client commands. All handlers must be of the form:
//   int handle_cmd_X(int argc, char **argv, int sock_fd)
// When called, argv will be a vector of strings, with the first string always
// being the command name, and argc will be the number of strings in the
// vector. This is exactly like argc and argv for main().

// Handler for get command
int handle_cmd_get(int argc, char **argv, int sock_fd) {
	char *msg;
	int ret;

	if (argc < MIN_GET_ARGS + 1)
		usage_and_exit(EXIT_FAILURE, "Not enough arguments for `get'");

	// Generate command string
	msg = (char*) malloc(MAX_MSG_LEN);  // Free me!
	if (!msg) { perror("handle_cmd_get"); exit(EXIT_FAILURE); }
	snprintf(msg, MAX_MSG_LEN, "GET %s %s %s",
	         argv[1], argv[2], argv[3]);
	ret = send_command(sock_fd, msg);
	free(msg);

	if (ret == DAP_SUCCESS)
		printf("GET command completed successfully!\n");
	else
		printf("GET command failed.\n");

	exit(EXIT_SUCCESS);
}

// Handler for list command
int handle_cmd_list(int argc, char **argv, int sock_fd) {
	if (argc < MIN_LIST_ARGS + 1)
		usage_and_exit(EXIT_FAILURE, "Not enough arguments for `list'");

	fprintf(stderr, "NOT YET IMPLEMENTED\n");
	exit(EXIT_FAILURE);
}

// MISC
// ----
// Functions and structures that handle miscellaneous tasks not related to
// client-server communication.

// Fork a new stork_log daemon, return the pid or exit.
int fork_stork_log_daemon() {
	int pid;
	
	switch(pid = fork()) {
		case -1:  // Something bad happened
		perror("fork_stork_log_daemon");
		exit(EXIT_FAILURE);

		case 0:  // We're the child process
		fclose(stdout);
		daemon_mode();
		exit(EXIT_FAILURE);  // If we get here, that's bad

		default:  // We're the parent process
		return pid;
	}
}

// Return string containing path name for pid file. Don't free it!
char *pid_path() {
	static char *path = NULL;

	// Allocate space and write the string if we haven't yet
	if (!path) {
		char *stork_dir = param("RELEASE_DIR");

		if (!stork_dir) {
			fprintf(stderr, "Warning: RELEASE_DIR variable not set.\n");
			stork_dir = "";
		}

		asprintf(&path, "%s/tmp/" PID_FILE_NAME, stork_dir);
		if (!path) { perror("pid_path"); exit(EXIT_FAILURE); }
	}

	return path;
}

// Return string containing the socket address. Multiple calls to this function
// with the same argument return the same pointer -- don't free it! Calls to
// this function with a different pid will free the old string, so don't rely
// on it always being there. strdup if necessary.
char *sock_path(int pid) {
	static int _pid = 0;
	static char *path = NULL;

	// Build string if we haven't yet or pid has changed
	if (_pid != pid || !path) {
		char *fmt_str = "%s/tmp/" SOCK_PREFIX "%d";
		char *stork_dir = param("RELEASE_DIR");
		size_t len;

		if (!stork_dir) {
			fprintf(stderr, "Warning: RELEASE_DIR variable not set.\n");
			stork_dir = "";
		}

		free(path);
		asprintf(&path, "%s/tmp/" SOCK_PREFIX "%d", stork_dir, _pid = pid);
		if (!path) { perror("sock_path"); exit(EXIT_FAILURE); }
	}

	return path;
}

// Pass no args to just use the current pid
inline char *sock_path() { return sock_path(getpid()); }

// Print usage information to a stream (generally either stderr or stdout)
void usage(FILE *fp) { fprintf(fp, USAGE); }

// Print usage and exit with given exit code. Prepend err_msg if given.
void usage_and_exit(int exit_code, char *err_msg) {
	// We don't want to print to stderr unless we're exiting badly
	FILE *fp = (!exit_code || exit_code == EXIT_SUCCESS) ? stdout : stderr;

	// Print err_msg if one was passed
	if (err_msg) fprintf(fp, "\n  Error: %s\n\n", err_msg);

	usage(fp);

	exit(exit_code);
}

// Struct representing parsed command line arguments
struct parsed_args_t {
	// Command handler
	int (*handler)(int, char**, int);

	int cmd_argc;     // Number of arguments in argument vector
	char **cmd_argv;  // Argument vector

	int pid;  // Process id of stork_log daemon
};

// Parse arguments and write results into passed parsed_args struct. If a
// command was given, it should assign a command handler. If an error occurs,
// print usage information and exit.
void parse_args(int argc, char **argv, parsed_args_t *pa) {
	// Set up parsed_args_t struct
	pa->pid = 0;
	pa->cmd_argv = NULL;
	pa->handler = NULL;

	// Parse out options. This isn't considered an evil use of goto, is it?
	for (int i = 1; i < argc; i++) {
		// First arg not starting with - must be the command!
		if (argv[i][0] != '-') {
			pa->cmd_argc = argc - i;
			pa->cmd_argv = &argv[i];
			break;
		}

		// See if it's one of those long arguments
		if (argv[i][1] == '-') {
			char *opt = &(argv[i][2]);
			if (!strcmp(opt, "help")) goto help;
			if (!strcmp(opt, "pid")) goto pid;
			goto invalid;
		}

		// Otherwise, if it's a single character argument, we can use a switch
		if (!argv[i][2]) switch (argv[i][1]) {
			case 'h': goto help;
			case 'p': goto pid;
			case 'q': goto quiet;
			case 'd': goto daemon_mode;
		}

		// Error handling
		invalid:  // If we get here, an invalid argument was passed
			usage_and_exit(EXIT_FAILURE, "Invalid option");
		dupe:  // Go here if there was a duplicate argument passed
			usage_and_exit(EXIT_FAILURE, "Duplicate argument detected");

		// Actually handle the various options
		help:  // Just display usage and exit successfully
			usage_and_exit(EXIT_SUCCESS, NULL);
		pid:  // The next argument should be the pid
			if (pa->pid) goto dupe;
			if (++i < argc) pa->pid = atoi(argv[i]);
			if (pa->pid <= 0)
				usage_and_exit(EXIT_FAILURE, "Invalid pid specified");
			continue;
		quiet:  // Don't print anything
			fclose(stdout);
			fclose(stderr);
			continue;
		daemon_mode:
			daemon_mode(0);
			exit(EXIT_SUCCESS);
	}

	// If we get here and don't know the command, we weren't given enough args
	if (!pa->cmd_argv) usage_and_exit(EXIT_FAILURE, "No command given");

	// Match command with its handler
	if (!strcmp(pa->cmd_argv[0], "get"))
		pa->handler = handle_cmd_get;
	else if (!strcmp(pa->cmd_argv[0], "list"))
		pa->handler = handle_cmd_list;
	else
		usage_and_exit(EXIT_FAILURE, "Invalid command");
}

// MAIN
// ----
int main(int argc, char **argv) {
	int sock_fd;
	parsed_args_t pa;

	// Load STORK_CONFIG
	config();

	// Parse arguments and exit if there's a usage problem
	parse_args(argc, argv, &pa);

	// Get PID from either parsed args or get_server_pid(), or fork otherwise
	if (!pa.pid && !(pa.pid = get_server_pid())) {
		printf("stork_log daemon does not appear to be running\n"
		       "Starting new daemon...\n");
		pa.pid = fork_stork_log_daemon();
		//wait_for_pid_file();
		sleep(1);
	}
	
	// Connect to the daemon and get a socket descriptor
	sock_fd = connect_to_daemon(pa.pid);
	
	// Now issue the command
	assert(pa.handler != NULL);
	pa.handler(pa.cmd_argc, pa.cmd_argv, sock_fd);
}
